#include "SPEEX\config.h"
#include "SPEEX\speex.h"
#include "SPEEX\speex_callbacks.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "types.h"
#include "wav_def.h"
#include "codec_interface.h"






//--------------------------

SPEEX_CODEC::SPEEX_CODEC()
{
spx_int32_t tmp;
spx_int32_t skip_group_delay;
int samplerate = SPEEX_SAMPLE_RATE;
       
// Set mode for encoder and decoder

#ifdef MODE_8K
  // Narrowband Fsa=8K
   st = speex_encoder_init(speex_lib_get_mode(SPEEX_MODEID_NB));
   dec = speex_decoder_init(speex_lib_get_mode(SPEEX_MODEID_NB));
#endif

#ifdef MODE_16K
  // Wideband Fsa=16K
   st = speex_encoder_init(speex_lib_get_mode(SPEEX_MODEID_WB));
   dec = speex_decoder_init(speex_lib_get_mode(SPEEX_MODEID_WB));
#endif


  // Use perceptual enhancement filter in decoder

   tmp=1;
   speex_decoder_ctl(dec, SPEEX_SET_ENH, &tmp);
   
   // Disable VBR

   tmp=0;
   speex_encoder_ctl(st, SPEEX_SET_VBR, &tmp);

   // Set quality (i.e. bitrate)
   // 

   tmp = QUALITY;
   speex_encoder_ctl(st, SPEEX_SET_QUALITY, &tmp);
   
   tmp = COMPLEXITY;
   speex_encoder_ctl(st, SPEEX_SET_COMPLEXITY, &tmp);
  
   // Enable HPF

   tmp=1;
   speex_encoder_ctl(st, SPEEX_SET_HIGHPASS, &tmp);
   speex_decoder_ctl(dec, SPEEX_SET_HIGHPASS, &tmp);

  
   speex_encoder_ctl(st, SPEEX_GET_LOOKAHEAD, &skip_group_delay);
   speex_decoder_ctl(dec, SPEEX_GET_LOOKAHEAD, &tmp);
   skip_group_delay += tmp;

speex_bits_init(&coder_bits);
speex_bits_init(&decoder_bits);

// EC and postprocessor

echo_state = speex_echo_state_init(FRAME_SIZE, EC_FILTER_LENGTH);
preprocess_state = speex_preprocess_state_init(FRAME_SIZE,  samplerate);

speex_echo_ctl(echo_state, SPEEX_ECHO_SET_SAMPLING_RATE, &samplerate);
speex_preprocess_ctl(preprocess_state, SPEEX_PREPROCESS_SET_ECHO_STATE, echo_state);

// Set AGC to default state

#ifdef AGC_ENABLED_BY_DEFAULT
agc_enable = 1;
tmp = 1;
#else
agc_enable = 0;
tmp = 0;
#endif

speex_preprocess_ctl(preprocess_state, SPEEX_PREPROCESS_SET_AGC, &tmp);


// Flat delay

flat_delay_buffer_samples = FLAT_DELAY_SAMPLES;
flat_delay_buffer = new s16[flat_delay_buffer_samples];
flat_delay_buffer_idx = 0;
memset(flat_delay_buffer, 0, sizeof(s16)*flat_delay_buffer_samples);

delayed_spkr_data = new s16[FRAME_SIZE];
memset(delayed_spkr_data, 0, sizeof(s16)*FRAME_SIZE);
}
//----------------------
// Destructor
//
SPEEX_CODEC::~SPEEX_CODEC()
{
delete[] delayed_spkr_data;
delete[] flat_delay_buffer;   
speex_preprocess_state_destroy(preprocess_state);    
speex_echo_state_destroy(echo_state);    
speex_bits_destroy(&decoder_bits);
speex_bits_destroy(&coder_bits);
speex_decoder_destroy(dec);
speex_encoder_destroy(st);
}
//------------------------------
void SPEEX_CODEC::Process_EC(s16 *output_audio_data, s16 *mic_audio_data, s16 *spkr_audio_data)
{
// Put speaker data into delay line

for(u32 ci = 0; ci < FRAME_SIZE; ci++)
 {
 delayed_spkr_data[ci] = flat_delay_buffer[flat_delay_buffer_idx];
 flat_delay_buffer[flat_delay_buffer_idx] = spkr_audio_data[ci];
 flat_delay_buffer_idx++;
 if(flat_delay_buffer_idx >= flat_delay_buffer_samples) flat_delay_buffer_idx = 0;
 }

speex_echo_cancellation(echo_state, mic_audio_data, delayed_spkr_data, output_audio_data);
speex_preprocess_run(preprocess_state, output_audio_data);
}

//-----------------------------------
// Returns the number of packed bytes
//
u32 SPEEX_CODEC::Pack_Audio_Data(u8 *buffer, s16 * input_audio_data)
{
speex_bits_reset(&coder_bits);

// Encode input frame into bits

speex_encode_int(st, input_audio_data, &coder_bits);

// Pack bits into packet

return speex_bits_write(&coder_bits, (char *)buffer, ENCODED_PACKET_MAX_SIZE_BYTES);
}
//----------------

void SPEEX_CODEC::Unpack_Audio_Data(s16 *output_audio_data, u8 *packed_buffer, u32 packed_data_bytes)
{

if((packed_buffer != NULL) && (packed_data_bytes != 0))
 {
 speex_bits_reset(&decoder_bits);   
 speex_bits_read_from(&decoder_bits, (const char *)packed_buffer, packed_data_bytes);
 speex_decode_int(dec, &decoder_bits, output_audio_data);
 }
else
 {
 speex_decode_int(dec, NULL, output_audio_data);
 }
}

void SPEEX_CODEC::Enable_AGC(int param) // 0 = disable, 1 = enable
{    
if(param != agc_enable)
 {
 spx_int32_t tmp = agc_enable = param;
 speex_preprocess_ctl(preprocess_state, SPEEX_PREPROCESS_SET_AGC, &tmp);
 }
}

int SPEEX_CODEC::AGC_Is_Enabled() // 0 = disable, 1 = enable
{
return agc_enable;
}



//======================= <END> =========================

