// Obfuscation of the data
//
#include "obfuscation.h"
#include "random.h"
#include <string.h>

static const u32 packet_max_size_bytes = 512;



static inline u8 obfuscation_prng(u32 &state)
{
state ^= 0x259f1ba9; // Constant is M-sequence of length 31 extended by 1
state = (state << 8) - state;
return (u8)(state >> 24);
}

u32 Obfuscate_Packet(u8 *dst, u8 const *src, const u32 src_len_bytes)
{
if(src_len_bytes > (packet_max_size_bytes - 2*sizeof(u32))) return 0;

u32 nonce;

cheap_random_number(&nonce, 1);

memcpy(dst, &nonce, sizeof(u32));
dst += sizeof(u32);

u8 *ptr = (u8*)(&src_len_bytes);
for(u32 i = 0; i < sizeof(u32); i++) *dst++ = obfuscation_prng(nonce) ^  (*ptr++);

for(u32 i = 0; i < src_len_bytes; i++) *dst++ = obfuscation_prng(nonce) ^  (*src++);

u32 padding_len_bytes = packet_max_size_bytes - src_len_bytes - 2*sizeof(u32);

if(padding_len_bytes) 
 {
 cheap_random_number(&nonce, 1); 
 padding_len_bytes = nonce % (padding_len_bytes + 1);
 }

for(u32 i = 0; i < padding_len_bytes; i++) *dst++ = obfuscation_prng(nonce); 

return 2*sizeof(u32) + src_len_bytes + padding_len_bytes;
}

u32 Deobfuscate_Packet(u8 *dst, u8 const *src, const u32 src_len_bytes)
{

if(src_len_bytes <= (2*sizeof(u32))) return 0;

if(src_len_bytes > packet_max_size_bytes) return 0;

u32 nonce, dst_len_bytes;

memcpy(&nonce, src, sizeof(u32));
src += sizeof(u32);

u8 *ptr = (u8*)(&dst_len_bytes);
for(u32 i = 0; i < sizeof(u32); i++) *ptr++ = obfuscation_prng(nonce) ^  (*src++);

if(dst_len_bytes > (src_len_bytes - 2*sizeof(u32))) return 0;

for(u32 i = 0; i < dst_len_bytes; i++) *dst++ = obfuscation_prng(nonce) ^ (*src++); 

return dst_len_bytes;
}
