// XEX counter mode
// data = 64 bit u32[4]
// key = 256 bit u32[8]
// cntr = 64 bit u32[4]

#include "xex_mode.h"
#include "xtea.h"
#include "idea.h"
#include "large_math.h"


void XEX_Encrypt(u32 *data, u32 const* key, u32 const* cntr)
{
u32 xor_data[2];

xor_data[0] = cntr[0];
xor_data[1] = cntr[1];

IDEA_Encode(xor_data, key);

data[0] ^= xor_data[0];
data[1] ^= xor_data[1];

XTEA_Encode(data, key + 4);

data[0] ^= xor_data[0];
data[1] ^= xor_data[1];

zero(xor_data, 2);
}

void XEX_Decrypt(u32 *data, u32 const* key, u32 const* cntr)
{
u32 xor_data[2];

xor_data[0] = cntr[0];
xor_data[1] = cntr[1];

IDEA_Encode(xor_data, key);

data[0] ^= xor_data[0];
data[1] ^= xor_data[1];

XTEA_Decode(data, key + 4);

data[0] ^= xor_data[0];
data[1] ^= xor_data[1];

zero(xor_data, 2);
}

// Length must be a multiple of 8 bytes

void XEX_Encrypt_Block(u32 *data, u32 *cntr,  u32 const *key, u32 length_x32)
{
for(u32 ci = 0; ci < length_x32; ci += 2)
 {
 XEX_Encrypt(data + ci, key, cntr);
 
 cntr[0]++;
 if(!cntr[0]) cntr[1]++;
 }
}

void XEX_Decrypt_Block(u32 *data, u32 *cntr,  u32 const *key, u32 length_x32)
{
for(u32 ci = 0; ci < length_x32; ci += 2)
 {
 XEX_Decrypt(data + ci, key, cntr);
 
 cntr[0]++;
 if(!cntr[0]) cntr[1]++;
 }
}

//========================= <END> ================================

