#include "xtea.h"


// XTEA/XXTEA (c) 1997-1998 A. Roger M. Needham and David J. Wheeler.
//
//

static const u32 xtea_rounds = 66;    // XTEA rounds (33 required to achieve full confusion): x2 Feistel rounds, so 33 = 66 Feistel rounds
                                      // Most efficient published attacks = 27 Feistel rounds diff. analysis and 36 rounds with related key
                                      // 66 XTEA rounds = 132 Feistel rounds                                     

static const u32 delta = 0x9E3779B9;  // XTEA internal value
static const u32 xxtea_min_rounds = 6; // Originally suggested 6 rounds are weak. Need 10 rounds at least


//----------------------------------------------------

void XTEA_Encode(u32* v, u32 const* k) 
{
u32 sum = 0;
u32 v0 = v[0];
u32 v1 = v[1];
u16 rounds = xtea_rounds;   

while(rounds--) 
 {
 v0  += (((v1 << 4) ^ (v1 >> 5)) + v1) ^ (sum + k[sum&3]);
 sum += delta;
 v1  += (((v0 << 4) ^ (v0 >> 5)) + v0) ^ (sum + k[(sum>>11)&3]);
 }

v[0] = v0; 
v[1] = v1;
}
//-------------

void XTEA_Decode(u32* v, u32 const* k) 
{
u32 sum=(u32)(((u64)delta)*xtea_rounds); // To avoid warning on u32 overflow
u32 v0 = v[0];
u32 v1 = v[1];
u16 rounds = xtea_rounds;

while(rounds--) 
 {
 v1  -= (((v0 << 4) ^ (v0 >> 5)) + v0) ^ (sum + k[(sum>>11) & 3]);
 sum -= delta;
 v0 -= (((v1 << 4) ^ (v1 >> 5)) + v1) ^ (sum + k[sum & 3]);
 }

v[0] = v0; 
v[1] = v1;
}

//---------------------
//
void Block_XXTEA_Encode(u32 *v, const u32 n, u32 const *k, const u32 rounds)
{
u32 e, p, q;

if(rounds) q = rounds;
else q = xxtea_min_rounds + 52/n; // weak!

u32 z = v[n-1];
u32 sum  = 0;

while(q--) 
 {
 sum += delta;
 e = (sum >> 2) & 3;
 
 for(p = 0; p < n; p++)
  {
  z = v[p] += (((z<<4) ^ (z>>5)) + z) ^ (k[(p&3)^e] + sum);
  }
 }
}

void Block_XXTEA_Decode(u32 *v, const u32 n, u32 const *k, const u32 rounds)
{
u32 e, p, q;

if(rounds) q = rounds;
else q = xxtea_min_rounds + 52/n;

u32 z = v[n-1];
u32 sum = q*delta;

while(sum) 
 {
 e = (sum >> 2) & 3;

 for(p = n-1; p > 0; p--) 
  {
  z = v[p-1];
  v[p] -= (((z<<4) ^ (z>>5)) + z) ^ (k[(p&3)^e] + sum);
  }
 
 z = v[n-1];
 v[0] -= (((z<<4) ^ (z>>5)) + z) ^ (k[(p&3)^e] + sum);
 sum -= delta; 
 }
}
//-----------------------------------------
// 
#define XXTEA_MX ((((z>>5)^(y<<2)) + ((y>>3)^(z<<4))) ^ ((sum ^ y) + (z ^ k[(p&3)^e])))



void XXTEA_Encode(u32 *v, const u32 n, u32 const *k, const u32 rounds)
{
u32 p;    
u32 q = (rounds) ? rounds : (xxtea_min_rounds + 52/n); 
u32 z = v[n-1];
u32 y = v[0];
u32 sum = 0;


while(q--) 
 {
 sum += delta;
 u32 e = (sum >> 2) & 3;

 for(p = 0; p < n-1 ; p++)
  {
  y = v[p+1];
  z = v[p] += XXTEA_MX;
  }

 y = v[0];
 z = v[n-1] += XXTEA_MX;
 }
}

void XXTEA_Decode(u32 *v, const u32 n, u32 const *k, const u32 rounds)
{
u32 p;    
u32 q = (rounds) ? rounds : (xxtea_min_rounds + 52/n); 
u32 z = v[n-1];
u32 y = v[0];
u32 sum = q*delta;

while(q--) 
 {
 u32 e = (sum >> 2) & 3;

 for(p = n-1; p > 0; p--)
  {
  z = v[p-1];
  y = v[p] -= XXTEA_MX;
  z = v[n-1];
  }
 y = v[0] -= XXTEA_MX;
 sum -= delta; 
 }
}

#undef XXTEA_MX 


#ifdef TEST_XTEA

// Block XXTEA from original article

#define MX (z>>5^y<<2)+(y>>3^z<<4)^ (sum ^y) +(k[ p&3^ e]^z ) ;

long btea( long * v, long n , long * k ) 
{
unsigned long z=v[n-1], y=v[0], sum=0,e,
DELTA=0x9e3779b9 ;
long m, p, q ;
if ( n>1) {
/* Coding Part */
q = 6+52/n ;
while ( q-- > 0 ) {
sum += DELTA ;
e = sum >> 2&3 ;
for ( p = 0 ; p < n-1 ; p++ )
y = v[p+1],
z = v[p] += MX
y = v[0] ;
z = v[n-1] += MX
}
return 0 ; }
/* Decoding Part */
else if ( n <-1 ) {
n = -n ;
q = 6+52/n ;
sum = q*DELTA ;
while (sum != 0) {
e = sum>>2 & 3 ;
for (p = n-1 ; p > 0 ; p-- )
z = v[p-1],
y = v[p] -= MX
z = v[n-1] ;
y = v[0] -= MX
sum -= DELTA ; }
return 0 ; }
return 1 ; } /* Signal n=0,1,-1 */


// Test functions
//
// block XTEA from original article
//
long teab( long * v, long n , long * k ) 
{
unsigned long z=v[n-1], sum=0,e,
DELTA=0x9e3779b9 ;
long m, p, q ;
if( n>1) 
 {
 /* Coding Part */
 q=6+52/n ;
 while (q-- > 0) 
 {
 sum += DELTA ;
 e = sum>>2&3 ;
 for (p=0 ; p<n ; p++ )
 z=v[p] += (z<<4 ^ z>>5) + z ^ k[p&3^e] + sum ;
 }
return 0 ; 
} 
/* Decoding Part */
else if (n<-1) 
{
n= -n ;
q=6+52/n ;
sum=q*DELTA ;
while (sum != 0) {
e= sum>>2 & 3 ;
for (p=n-1 ; p>0 ; p-- ) {
z=v[p-1] ;
v[p] -= (z<<4 ^ z>>5) + z ^ k[p&3^e] + sum ;
}
z=v[n-1] ;
v[0] -= (z<<4 ^ z>>5) + z ^ k[p&3^e] + sum ;
sum-=DELTA ; }
return 0 ; }

return 1 ; 
} /* Signal n=0 */




//----------------------------------
// XTEA from original article
//
void tean(long * v, long * k, long N) 
{
unsigned long y=v[0], z=v[1], DELTA=0x9e3779b9;

if(N>0) 
{
// coding 
unsigned long limit=DELTA*N, sum=0 ;
while(sum!=limit)
y+= (z<<4 ^ z>>5) + z ^ sum + k[sum&3],
sum+=DELTA,
z+= (y<<4 ^ y>>5) + y ^ sum + k[sum>>11 &3] ;
}
else 
{
// decoding
unsigned long sum=DELTA*(-N) ;
while (sum)
z-= (y<<4 ^ y>>5) + y ^ sum + k[sum>>11 &3],
sum-=DELTA,
y-= (z<<4 ^ z>>5) + z ^ sum + k[sum&3] ;
}
v[0]=y, v[1]=z;
}

#endif

//================= <END> ==============
