#include <tchar.h>
#include "control.h"
#include "resource.h"
#include "main.h"
//----------------------------------------------------------------
CONTR_PROT::CONTR_PROT(NODE* node, HWND mWnd, HWND parent)
{
	mainWnd = mWnd;
	this->parent = parent;
    nd = *node;
    state = CTRL_ST_IDLE;
    msg = NULL;
    send_msg_num = 0;
    rec_msg_num = 0;
	rec_msg_flag = 0;
	first_msg_rcv = false;
    for (int i = 0; i < CONTR_MAX_VOICE_BUF; i++)
	{
        rec_data[i] = new u8[1024];
        rec_size[i] = 0;
	}
	rec_wr_inx = 0;
	rec_rd_inx = 0;
	rec_pack_num = 0;
	send_pack_num = 0;
	rec_data_buf_limit = 8;
	send_voice_buf = NULL;
	voice_cnt = 0;
	total_pack_num = 0;
	voice_pack_num = 0;
	voice_pack_proc = 0;
	InitializeCriticalSection(&cs);
	type_ticks = GetTickCount();
}
//----------------------------------------------------------------
CONTR_PROT::~CONTR_PROT()
{
	if (send_voice_buf) delete[] send_voice_buf;
    if (msg)
    {
        MESSAGE* cur;
        for (cur=msg;cur;)
        {
            MESSAGE* next = cur->next;
			delete cur->text;
            delete cur;
            cur = next;
        }
    }
	for (int i = 0; i < CONTR_MAX_VOICE_BUF; i++)
	{
		delete[] rec_data[i];
	}
	DeleteCriticalSection(&cs);
}
//----------------------------------------------------------------
static void TransferMessage(HWND hWnd, MESSAGE* message,NODE* nd)
{
	size_t len = _tcslen(message->text)*sizeof(TCHAR);
	u32 buf[NETWORK_SIZE_TEXT/4+2] = {0};
    buf[0] = CMD_SHOW_MESSAGE;
    buf[1] = message->msg_num;
	if (len > NETWORK_SIZE_TEXT) len = NETWORK_SIZE_TEXT;
	CopyMemory(&buf[2],message->text,len);
	if (len&0x1F) len = (len+0x20)&~0x1F;
	PACKET_FOR_SEND pack = {0};
	pack.buffer = buf;
	pack.buf_len = (u32)(2*sizeof(u32) + len);
	pack.nd = nd;
	pack.protocol = NETWORK_PROT_CONTROL;
	getNet()->send_packet(&pack);
    message->send_cnt++;
}
//----------------------------------------------------------------
void CONTR_PROT::sendMessage(TCHAR* message,LONG len,LONG posMin, LONG posMax)
{
    MESSAGE* ptr;
    if (state == CTRL_ST_IDLE) {
		state = CTRL_ST_CONNECTION;
		process();
        send_msg_num = 0;
        rec_msg_num = 0;
		PostMessage(parent,WM_UPDATECHAT,0,0);
	}
    if (msg)
    {
        for (ptr = msg; ptr->next; ptr = ptr->next);
        ptr->next = new MESSAGE;
        ptr = ptr->next;
    }
    else
    {
        msg = new MESSAGE;
        ptr = msg;
    }
    ptr->msg_num = send_msg_num++;
    ptr->next = NULL;
    ptr->send_cnt = 0;
    ptr->text = new TCHAR[len+1];
	_tcsncpy_s(ptr->text,len+1,message,len);
	ptr->min_pos = posMin;
	ptr->max_pos = posMax;

    TransferMessage(mainWnd,ptr,&nd);
}
//----------------------------------------------------------------
void CONTR_PROT::sendVoice(u8* data, u32 size)
{
	PACKET_FOR_SEND pack = {0};
	if (state != CTRL_ST_IN_VOICE) return;
	if (!send_voice_buf) 
	{
		send_voice_buf = new u32[size+1];
	}
	send_voice_buf[0] = CMD_VOICE;
	CopyMemory(&send_voice_buf[1],data,size);
	pack.buffer = send_voice_buf;
	pack.buf_len = sizeof(u32) + size;
	pack.nd = &nd;
	pack.protocol = NETWORK_PROT_CONTROL;
	pack.dont_count_tx = true;
	getNet()->send_packet(&pack);
}
//----------------------------------------------------------------
void CONTR_PROT::processMessages(void)
{
    MESSAGE* ptr;
    for (ptr = msg; ptr; ptr = ptr->next)
    {
        TransferMessage(parent,ptr,&nd);
    }
}

//----------------------------------------------------------------


void CONTR_PROT::process(void)
{
	PACKET_FOR_SEND pack = {0};
	pack.protocol = NETWORK_PROT_CONTROL;
	pack.nd = &nd;
	NETWORK* net = getNet();

    switch(state)
    {
    case CTRL_ST_IDLE:
    default:
		voice_cnt = 0;
        break;
    case CTRL_ST_CONNECTION:
		pack.buffer = &CMD_START_CONNECTION;
		pack.buf_len = sizeof(CMD_START_CONNECTION);
		if (net) net->send_packet(&pack);
        break;
    case CTRL_ST_CONNECTED:
    case CTRL_ST_ABORT:
    case CTRL_ST_DIALING:
    case CTRL_ST_IN_VOICE:
        processMessages();
        break;
    }
	if (state == CTRL_ST_ABORT)
	{
        voice_cnt++;
        if (voice_cnt > 3)
        {
            state = CTRL_ST_IDLE;
			PostMessage(parent,WM_UPDATECHAT,0,0);
        }
	}
    if (state == CTRL_ST_DIALING)
	{
        voice_cnt++;
		pack.buffer = &CMD_HANG_UP;
		pack.buf_len = sizeof(CMD_HANG_UP);
		if (net) net->send_packet(&pack);
        if (voice_cnt > 60)
        {
            state = CTRL_ST_IDLE;
			PostMessage(parent,WM_UPDATECHAT,0,0);
        }
	}
    if (state == CTRL_ST_IN_VOICE)
    {
        voice_cnt++;
		pack.buffer = &CMD_IN_VOICE;
		pack.buf_len = sizeof(CMD_IN_VOICE);
		if (net) net->send_packet(&pack);
        if (voice_cnt > 20)
        {
            state = CTRL_ST_DIALING;
			PostMessage(parent,WM_UPDATECHAT,0,0);
        }
    }
}

void CONTR_PROT::received(u32* packet, u32 size)
{
 	PACKET_FOR_SEND pack = {0};
	pack.protocol = NETWORK_PROT_CONTROL;
	pack.nd = &nd;
	NETWORK* net = getNet();

	total_pack_num++;

   if (size < sizeof(u32)) return;
	voice_cnt = 0;
    switch(packet[0]&0xFF) // command
    {
    case CMD_START_CONNECTION:
		pack.buffer = &CMD_CONNECTED;
		pack.buf_len = sizeof(CMD_CONNECTED);
		if (net) net->send_packet(&pack);
        if (state != CTRL_ST_CONNECTED)
        {
            state = CTRL_ST_CONNECTED;
            send_msg_num = 0;
            rec_msg_num = 0;
			PostMessage(parent,WM_UPDATECHAT,0,0);
        }
        break;
    case CMD_CONNECTED:
        if (state < CTRL_ST_CONNECTED)
        {
            state = CTRL_ST_CONNECTED;
			PostMessage(parent,WM_UPDATECHAT,0,0);
        }
        break;
    case CMD_SHOW_MESSAGE:
        if(size > sizeof(u32)*2)
        {
			// send responce
            u32 data[2] = {CMD_MESSAGE_RECEIVED,packet[1]};
			pack.buffer = data;
			pack.buf_len = sizeof(data);
			if (net) net->send_packet(&pack);
			// parse packet
			u32 cur_num = packet[1];
			if (cur_num > rec_msg_num)
            {
				SendMessageTimeout(parent,WM_SHOWTEXT,(WPARAM)&packet[2],size-sizeof(u32)*2,0,100,NULL);
				rec_msg_flag >>= cur_num - rec_msg_num;
                rec_msg_num = cur_num;
				rec_msg_flag |= 0x80000000;
				first_msg_rcv = false;
            }
			else if ((rec_msg_num && !cur_num) || ((rec_msg_num - cur_num) > 32) || (!rec_msg_num && !cur_num && !first_msg_rcv)) // session restart
			{
				SendMessageTimeout(parent,WM_SHOWTEXT,(WPARAM)&packet[2],size-sizeof(u32)*2,0,100,NULL);
				rec_msg_num = cur_num;
				rec_msg_flag = 0x80000000;
				first_msg_rcv = true;
			}
			else
			{
				u8 shift = rec_msg_num - cur_num;
				u32 mask = 0x80000000 >> shift;
				if (!(rec_msg_flag & mask))
				{
					SendMessageTimeout(parent,WM_SHOWTEXT,(WPARAM)&packet[2],size-sizeof(u32)*2,0,100,NULL);
				}
				rec_msg_flag |= mask;
			}
        }
        break;
    case CMD_MESSAGE_RECEIVED:
        if(size >= sizeof(u32)*2)
        {
            MESSAGE *ptr, *prev_ptr;
            for (prev_ptr = ptr = msg; ptr; ptr = ptr->next)
            {
                if (ptr->msg_num == packet[1])
                {
                    if (prev_ptr == ptr) // very first item
                    {
                        msg = ptr->next;
                    }
                    else
                    {
                        prev_ptr->next = ptr->next;
                    }
					delete ptr->text;
					PostMessage(parent,WM_MESSAGE_DELIVERED,ptr->min_pos,ptr->max_pos);
                    delete ptr;
                    break;
                }
            }
        }
        break;
    case CMD_HANG_UP:
		switch (state)
		{
		case CTRL_ST_ABORT:
			pack.buffer = &CMD_HANG_DOWN;
			pack.buf_len = sizeof(CMD_HANG_DOWN);
			if (net) net->send_packet(&pack);
			break;
		case CTRL_ST_DIALING:
			state = CTRL_ST_IN_VOICE;
		case CTRL_ST_IN_VOICE:
			pack.buffer = &CMD_IN_VOICE;
			pack.buf_len = sizeof(CMD_IN_VOICE);
			if (net) net->send_packet(&pack);
			break;
		case CTRL_ST_IDLE:
			state = CTRL_ST_CONNECTED;
		default:
			PostMessage(parent,WM_RING,0,0);
			break;
		}
        break;
    case CMD_HANG_DOWN:
        state = CTRL_ST_IDLE;
		PostMessage(parent,WM_UPDATECHAT,0,0);
        break;
    case CMD_IN_VOICE:
        if (state == CTRL_ST_DIALING)
		{
            state = CTRL_ST_IN_VOICE;
			voice_buf_filled = false;
			PostMessage(parent,WM_UPDATECHAT,0,0);
		}
        else if (state < CTRL_ST_IN_VOICE)
		{
			pack.buffer = &CMD_HANG_DOWN;
			pack.buf_len = sizeof(CMD_HANG_DOWN);
			if (net) net->send_packet(&pack);
		}
        break;
    case CMD_VOICE:
		voice_pack_num++;
        if (state == CTRL_ST_IN_VOICE)
        {
			int wr_inx;
			EnterCriticalSection(&cs);
			wr_inx = rec_wr_inx + 1;
			if (wr_inx >= CONTR_MAX_VOICE_BUF) wr_inx -= CONTR_MAX_VOICE_BUF;
			if (wr_inx != rec_rd_inx)
			{
				rec_size[rec_wr_inx] = size-sizeof(u32);
				if (rec_size[rec_wr_inx] < 1024)
				{
					memcpy(rec_data[rec_wr_inx],&packet[1],rec_size[rec_wr_inx]);
					rec_wr_inx = wr_inx;
				}
				else
				{
					rec_size[rec_wr_inx] = 0;
				}
			}
			wr_inx = rec_wr_inx - rec_rd_inx;
			if (wr_inx < 0) wr_inx += CONTR_MAX_VOICE_BUF;
			if (wr_inx >= rec_data_buf_limit)
			{
				voice_buf_filled = true;
			}
			LeaveCriticalSection(&cs);
        }
        break;
    case CMD_TYPING:
		PostMessage(parent,WM_SHOW_TYPING,0,0);
        break;
	case CMD_REJECT:
		state = CTRL_ST_REJECTED;
		PostMessage(parent,WM_UPDATECHAT,0,0);
		break;
	}
}

void CONTR_PROT::accept(void)
{
    state = CTRL_ST_IN_VOICE;
	voice_buf_filled = false;
    voice_cnt = 0;
	PostMessage(parent,WM_UPDATECHAT,0,0);
}

void CONTR_PROT::reject(void)
{
	NETWORK* net = getNet();
 	PACKET_FOR_SEND pack = {0};
	pack.protocol = NETWORK_PROT_CONTROL;
	pack.nd = &nd;

    voice_cnt = 0;
    state = CTRL_ST_ABORT;
	pack.buffer = &CMD_HANG_DOWN;
	pack.buf_len = sizeof(CMD_HANG_DOWN);
	if (net) net->send_packet(&pack);
	PostMessage(parent,WM_UPDATECHAT,0,0);
}

void CONTR_PROT::dial(void)
{
    if (state < CTRL_ST_DIALING)
	{
		NETWORK* net = getNet();
 		PACKET_FOR_SEND pack = {0};
		pack.protocol = NETWORK_PROT_CONTROL;
		pack.nd = &nd;

	    voice_cnt = 0;
		state = CTRL_ST_DIALING;
		pack.buffer = &CMD_HANG_UP;
		pack.buf_len = sizeof(CMD_HANG_UP);
		if (net) net->send_packet(&pack);
		PostMessage(parent,WM_UPDATECHAT,0,0);
	}
}

void CONTR_PROT::hang_down(void)
{
    if (state > CTRL_ST_CONNECTED)
	{
		NETWORK* net = getNet();
 		PACKET_FOR_SEND pack = {0};
		pack.protocol = NETWORK_PROT_CONTROL;
		pack.nd = &nd;

		state = CTRL_ST_IDLE;
		pack.buffer = &CMD_HANG_DOWN;
		pack.buf_len = sizeof(CMD_HANG_DOWN);
		if (net) net->send_packet(&pack);
		PostMessage(parent,WM_UPDATECHAT,0,0);
	}
}

u8* CONTR_PROT::getVoice(u32* size)
{
	u8* buf = NULL;
	*size = 0;
    if ((state != CTRL_ST_IN_VOICE) || !voice_buf_filled) return NULL;
	EnterCriticalSection(&cs);
	if (rec_rd_inx!=rec_wr_inx) // FIFO contains data
	{
		if (rec_size[rec_rd_inx])
		{
			buf = rec_data[rec_rd_inx];
			*size = rec_size[rec_rd_inx];
			rec_size[rec_rd_inx] = 0;
			rec_rd_inx++;
			if (rec_rd_inx >= CONTR_MAX_VOICE_BUF) rec_rd_inx = 0;
		}
		else
		{
			voice_buf_filled = false; // something goes wrong
		}
	}
	else // FIFO underflow
	{
		voice_buf_filled = false;
		if (rec_data_buf_limit < CONTR_MAX_VOICE_BUF-8)
		{
			rec_data_buf_limit+=8;
		}
	}
	LeaveCriticalSection(&cs);
	return buf;
}

void CONTR_PROT::sendTyping(void)
{
	DWORD ticks = GetTickCount();
	if (ticks - type_ticks > 2000)
	{
		NETWORK* net = getNet();
 		PACKET_FOR_SEND pack = {0};
		pack.protocol = NETWORK_PROT_CONTROL;
		pack.nd = &nd;
		pack.buffer = &CMD_TYPING;
		pack.buf_len = sizeof(CMD_TYPING);
		if (net) net->send_packet(&pack);

		type_ticks = ticks;
	}
}

void CONTR_PROT::getVoiceFifo(int* len, int* mean)
{
	if (state == CTRL_ST_IN_VOICE)
	{
		int diff = rec_wr_inx - rec_rd_inx;
		if (diff < 0) diff += CONTR_MAX_VOICE_BUF;
		*len = diff;
	}
	else
	{
		*len = 0;
	}
	*mean = rec_data_buf_limit;
}