#ifndef USERS_H_INCLUDED
#define USERS_H_INCLUDED
//-----------------------------------------------------------------------
#include "cp_threads.h"
#include "cp_strop.h"
#include "pugixml.h"
#include "kademlia.h"
#include "cp_time.h"
#include "types.h"
using namespace pugi;
//-----------------------------------------------------------------------
static const u32 ID_LENGTH_CCH = NETWORK_ID_LENGTH*(8+1);
static const u32 IP_LENGTH_CCH = 4*(3+1);
static const u32 TIME_LENGTH_CCH = 6*(2+1);
//-----------------------------------------------------------------------
typedef struct user {
	NODE nd;
	cp_time_t last_seen;
	ptstr_t url;
	struct user *next;
} user_t;
//-----------------------------------------------------------------------
typedef struct like {
	NODE nd;
	ptstr_t name;
	ptstr_t name_orig;
	bool trust;
	struct like *next;
} like_t;
//-----------------------------------------------------------------------
typedef struct download {
	NODE nd;
	ptstr_t file;
	ptstr_t user;
	cp_time_t time;
	struct download *next;
} download_t;
//-----------------------------------------------------------------------
class KADEMLIA;
//-----------------------------------------------------------------------
class USERS {
	lock_t lock;
	user_t* users;
	like_t* likes;
	download_t* downloads;

	void _append_user(user_t *new_user);

public:
	USERS();
	~USERS();

	void update_node(const NODE* nd);
	void delete_node(const NODE* nd);
	void load_nodes(KADEMLIA* kad, pctstr_t filename);
	void save_nodes(pctstr_t filename);

	bool load_likes(pctstr_t filename);
	void save_likes(pctstr_t filename);
	void append_like(NODE* nd, pctstr_t name, pctstr_t orig_name, bool trust);
	void remove_like(u32* id);
	void unauthorize_likes(void);
	like_t* getLikes(void) {return likes;};

	void load_downloads(pctstr_t filename);
	void save_downloads(pctstr_t filename);
	void append_download(NODE* nd, pctstr_t file, pctstr_t user, cp_time_t time = 0);
	void remove_download(u32 num);
	download_t* get_downloads(void) {return downloads;};
};
//-----------------------------------------------------------------------
void id2str(const u32 *id, ptstr_t str);
bool str2id(pctstr_t str, u32 *id);
void ip2str(u32 ip, ptstr_t str);
bool str2ip(pctstr_t str, u32 *ip);
void ip_port2str(u32 ip, u16 port, ptstr_t str);
void time2str(const cp_time_t t, ptstr_t str);
bool str2time(pctstr_t str, cp_time_t *t);
//-----------------------------------------------------------------------
#endif //USERS_H_INCLUDED
